program Lesson25_Morphing;

// This code was created by Jeff Molofee '99
// If you've found this code useful, please let me know.
// Visit me at www.demonews.com/hosted/nehe

{ Translation to Delphi/Object Pascal by Alexandre Hirzel, 2003
                    (Alexandre.Hirzel@ie-zea.unil.ch)
}

uses
  Windows,
  Messages,
  OpenGL,
  SysUtils; // Need StrToInt()

{$R *.RES}

type
  tVertex=record
    x,y,z	:GLfloat;
  end;//record

  tObjet=class(tObject)
  public
    verts	:integer;           // Number of vertices
    Points	:array of tVertex;  // List of vertices
    constructor Create;overload;
    constructor Create(const FileName:string);overload;
    destructor Destroy;override;
    function LoadFromFile(const Filename:string):boolean;
  end;//record

var
  h_RC	: HGLRC;	   // Permanent Rendering Context
  h_DC	: HDC; 	   // Private GDI Device Context
  keys	: array [0..255] of BOOL;		// Array Used For The Keyboard Routine

  active : bool	= TRUE;	// Window Active Flag Set To TRUE By Default
  fullscreen : bool = TRUE;	// Fullscreen Flag Set To Fullscreen Mode By Default

  xrot,yrot,zrot 		:GLfloat; // Rotations
  cx,cy				:GLfloat;  // Coordinates
  cz				:GLfloat=-15;
  Sour,Dest,Helper				:tObjet;
  morph1,morph2,morph3,morph4	:tObjet;
  key	:integer=1;         // Used to make sure same morph key is not pressed
  Step   :integer=0;        	// Step counter
  Steps	:integer=200;      	// Maximum number of steps
  Morph	:boolean=false;     // Default morph to false (Not morphing)

function Calculate(const i:integer):tVertex;
begin
  Result.x:=(Sour.Points[i].x-Dest.Points[i].x)/steps;
  Result.y:=(Sour.Points[i].y-Dest.Points[i].y)/steps;
  Result.z:=(Sour.Points[i].z-Dest.Points[i].z)/steps;
end;

procedure InitGL(Width: GLsizei; Height: GLsizei); // This Will Be Called Right After The GL Window Is Created
var
  i	:integer;
begin
  glShadeModel(GL_SMOOTH);						// Enable Smooth Shading
  glBlendFunc(GL_SRC_ALPHA,GL_ONE);				// Set the blending function for translucency
  glClearColor(0.0, 0.0, 0.0, 0.0);				// Black Background
  glClearDepth(1.0);							// Depth Buffer Setup
  glEnable(GL_DEPTH_TEST);						// Enables Depth Testing
  glDepthFunc(GL_LEQUAL);						// The Type Of Depth Testing To Do
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);	// Really Nice Perspective Calculations

  Morph1:=tObjet.Create('sphere.txt');  // Create the object and load data from file
  Morph2:=tObjet.Create('torus.txt');
  Morph3:=tObjet.Create('tube.txt');
  Morph4:=tObjet.Create;
  SetLength(Morph4.Points,486); // Manually reserver ram for a 4th 486 vertice object (morph4)
  Morph4.verts:=486;
  for i:=0 to 485 do with Morph4.Points[i] do begin  // Loop through all 486 vertices
    x:=random*14-7;     // morph4 x point becomes a random float value from -7 to 7
    y:=random*14-7;     // morph4 y point becomes a random float value from -7 to 7
    z:=random*14-7;     // morph4 z point becomes a random float value from -7 to 7
  end;//for i

  Helper:=tObjet.Create('sphere.txt');    // Load sphere.txt object into helper (Used as starting point)
  Sour:=Morph1;      //
  Dest:=Morph1;      // Source & destination are set to equal first object (morph1)
end;

procedure ReSizeGLScene(Width: GLsizei; Height: GLsizei);
var
  fWidth, fHeight: GLfloat;
begin
  if (Height=0) then Height:=1;   // Prevent A Divide By Zero If The Window Is Too Small

  glViewport(0, 0, Width, Height);// Reset The Current Viewport And Perspective Transformation

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();

  fWidth := Width;
  fHeight := Height;
  gluPerspective(45.0,fWidth/fHeight,0.1,100.0);
  glMatrixMode(GL_MODELVIEW);
end;

procedure DrawGLScene();
var
  tx,ty,tz	:GLfloat; // Temporary coordinates
  q			:tVertex; // Holds returned calculated values for one vertex
  i			:integer;
begin
  glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT); // Clear The Screen And The Depth Buffer
  glLoadIdentity();							 // Reset The View

  glTranslatef(cx,cy,cz);		// Translate the current position to start drawing
  glRotatef(xrot,1,0,0);		// Rotate on the X axis by xrot
  glRotatef(yrot,0,1,0);		// Rotate on the Y axis by yrot
  glRotatef(zrot,0,0,1);		// Rotate on the Z axis by zrot

  glBegin(GL_POINTS);		// Begin drawing points
    for i:=0 to Morph1.verts-1 do begin // Loop through all the verts of morph1 (All objects have the same amount of verts for simplicity, could use maxcer also)
      if Morph then q:=Calculate(i)     // If morph is true calculate movement...
      else begin                        // ... otherwise movement = 0
        q.x:=0;
        q.y:=0;
        q.z:=0;
      end;//else
      with Helper.Points[i] do begin
        x:=x-q.x;			// Subtract q.x units from helper.points[i].x (Move on X axis)
        y:=y-q.y;			// Subtract q.y units from helper.points[i].y (Move on Y axis)
        z:=z-q.z;			// Subtract q.z units from helper.points[i].z (Move on Z axis)
        tx:=x;			// Make Temp X variable equal to helper's X variable
        ty:=y;			// Make Temp Y variable equal to helper's Y variable
        tz:=z;			// Make Temp Z variable equal to helper's Z variable
      end;//with

      glColor3f(0,1,1);		     	// Set color to a bright shade of blue
      glVertex3f(tx,ty,tz);			// Draw a point at the current temp values (Vertex)
      glColor3f(0,0.5,1);			// Darken color a bit
      tx:=tx-2*q.x;
      ty:=ty-2*q.y;                    // Calculate two positions ahead
      tz:=tz-2*q.z;				
      glVertex3f(tx,ty,tz);			// Draw a second point at the newly calculate position
      glColor3f(0,0,1);		       	// Set color to a very dark blue
      tx:=tx-2*q.x;
      ty:=ty-2*q.y;                    // Calculate two positions ahead
      tz:=tz-2*q.z;
      glVertex3f(tx,ty,tz);			// Draw a third point at the second new position
    end;//for                          // This creates a ghostly tail as points move
  glEnd();	                         // Done drawing points

  // If we're morphing and we haven't gone through all 200 steps increase our step counter
  // Otherwise set morphing to false, make source = destination and set the step counter back to zero.
  if morph and (step <= steps) then Inc(step)
  else begin
    morph:= false;
    sour:= dest;
    step:= 0;
  end;//else
end;

function WndProc(hWnd: HWND; message: UINT;
                 wParam: WPARAM; lParam: LPARAM): LRESULT; stdcall;
var
  Screen		: TRECT;
  PixelFormat	: GLuint;
const
  pfd: PIXELFORMATDESCRIPTOR = (
  nSize: sizeof( PIXELFORMATDESCRIPTOR ); // Size Of This Pixel Format Descriptor
  nVersion: 1;                // Version Number (?)
  dwFlags: PFD_DRAW_TO_WINDOW // Format Must Support Window
    or PFD_SUPPORT_OPENGL     // Format Must Support OpenGL
    or PFD_DOUBLEBUFFER;      // Must Support Double Buffering
  iPixelType: PFD_TYPE_RGBA;  // Request An RGBA Format
  cColorBits: 16;             // Select A 16Bit Color Depth
  cRedBits: 0;                // Color Bits Ignored (?)
  cRedShift: 0;
  cGreenBits: 0;
  cBlueBits: 0;
  cBlueShift: 0;
  cAlphaBits: 0;              // No Alpha Buffer
  cAlphaShift: 0;             // Shift Bit Ignored (?)
  cAccumBits: 0;              // No Accumulation Buffer
  cAccumRedBits: 0;           // Accumulation Bits Ignored (?)
  cAccumGreenBits: 0;
  cAccumBlueBits: 0;
  cAccumAlphaBits: 0;
  cDepthBits: 16;             // 16Bit Z-Buffer (Depth Buffer)
  cStencilBits: 0;            // No Stencil Buffer
  cAuxBuffers: 0;             // No Auxiliary Buffer (?)
  iLayerType: PFD_MAIN_PLANE; // Main Drawing Layer
  bReserved: 0;               // Reserved (?)
  dwLayerMask: 0;             // Layer Masks Ignored (?)
  dwVisibleMask: 0;
  dwDamageMask: 0 );
begin
  Result := 0;
  case (message) of     // Tells Windows We Want To Check The Message
  WM_CREATE:
    begin
    h_DC := GetDC(hWnd);  // Gets a Device Context For The Window
    PixelFormat := ChoosePixelFormat(h_DC, @pfd);  // Finds The Closest Match To The Pixel Format We Set Above

    if (PixelFormat=0) then
      begin
      MessageBox(0,'Cant''t Find A Suitable PixelFormat.','Error',MB_OK or MB_ICONERROR);
      PostQuitMessage(0);  // This Sends A 'Message' Telling The Program To Quit
      exit;
      end;

    if (not SetPixelFormat(h_DC,PixelFormat,@pfd)) then
      begin
      MessageBox(0,'Can''t Set The PixelFormat.','Error',MB_OK or MB_ICONERROR);
      PostQuitMessage(0);
      exit;
      end;

    h_RC := wglCreateContext(h_DC);
    if (h_RC=0) then
      begin
      MessageBox(0,'Can''t Create A GL Rendering Context.','Error',MB_OK or MB_ICONERROR);
      PostQuitMessage(0);
      exit;
      end;

    if (not wglMakeCurrent(h_DC, h_RC)) then
      begin
      MessageBox(0,'Can''t activate GLRC.','Error',MB_OK or MB_ICONERROR);
      PostQuitMessage(0);
      exit;
      end;

    GetClientRect(hWnd, Screen);
    InitGL(Screen.right, Screen.bottom);
    end;
  WM_DESTROY, WM_CLOSE:
    begin
      ChangeDisplaySettings(DEVMODE(nil^), 0);

      Morph1.Free;
      Morph2.Free;
      Morph3.Free;
      Morph4.Free;

      wglMakeCurrent(h_DC,0);
      wglDeleteContext(h_RC);
      ReleaseDC(hWnd,h_DC);

      PostQuitMessage(0);
    end;
  WM_KEYDOWN:
    begin
      keys[wParam] := TRUE;
    end;
  WM_KEYUP:
    begin
      keys[wParam] := FALSE;
    end;
  WM_SIZE:
    begin
      ReSizeGLScene(LOWORD(lParam),HIWORD(lParam));
    end;
  else
    begin
      Result := DefWindowProc(hWnd, message, wParam, lParam);
      exit;
    end;
  end;//case
end;

procedure ProcessKeys;
begin
  if (keys[VK_PRIOR]) then zrot := zrot + 1;
  if (keys[VK_NEXT]) then zrot := zrot - 1;
  if (keys[VK_UP]) then xrot:=xrot-1;
  if (keys[VK_DOWN]) then xrot:=xrot+1;
  if (keys[VK_RIGHT]) then yrot:=yrot+1;
  if (keys[VK_LEFT]) then yrot:=yrot-1;
  if (keys[ord('Q')]) then cz:=cz-0.01;
  if (keys[ord('Z')]) then cz:=cz+0.01;
  if (keys[ord('W')]) then cy:=cy+0.01;
  if (keys[ord('S')]) then cy:=cy-0.01;
  if (keys[ord('D')]) then cx:=cx+0.01;
  if (keys[ord('A')]) then cx:=cx-0.01;
  if (keys[ord('1')])and(key<>1) and not Morph then begin
    key:=1;
    Morph:=True;
    Dest:=Morph1;
  end;//if
  if (keys[ord('2')])and(key<>2) and not Morph then begin
    key:=2;
    Morph:=True;
    Dest:=Morph2;
  end;//if
  if (keys[ord('3')])and(key<>3) and not Morph then begin
    key:=3;
    Morph:=True;
    Dest:=Morph3;
  end;//if
  if (keys[ord('4')])and(key<>4) and not Morph then begin
    key:=4;
    Morph:=True;
    Dest:=Morph4;
  end;//if
end;

function WinMain(	hInstance: HINST;
			hPrevInstance: HINST;
			lpCmdLine: PChar;
			nCmdShow: integer): integer; stdcall;
var
    msg: TMsg;      // Windows Message Structure
    wc: TWndClass;  // Windows Class Structure Used To Set Up The Type Of Window
    h_Wnd: HWND;     // Storage For Window Handle
	  dmScreenSettings: DEVMODE;
begin
  ZeroMemory( @wc, sizeof( wc ) );

  wc.style		  := CS_HREDRAW or CS_VREDRAW or CS_OWNDC;
  wc.lpfnWndProc   := @WndProc;
  wc.hInstance	  := hInstance;
  wc.hCursor	  := LoadCursor(0, IDC_ARROW);
  wc.lpszClassName := 'OpenGL WinClass';

  if(RegisterClass(wc)=0) then
    begin
       MessageBox(0,'Failed To Register The Window Class.','Error',MB_OK or MB_ICONERROR);
       Result := 0;
  exit;
    end;

  h_Wnd := CreateWindow(
  'OpenGL WinClass',
  'Jeff Molofee''s GL Code Tutorial ... NeHe ''99',		// Title Appearing At The Top Of The Window

  WS_POPUP or
  WS_CLIPCHILDREN or
  WS_CLIPSIBLINGS,

  0, 0,												  // The Position Of The Window On The Screen
  640, 480, 		// The Width And Height Of The WIndow

  0,
  0,
  hInstance,
  nil);

  if (h_Wnd = 0) then
    begin
       MessageBox(0,'Window Creation Error.','Error',MB_OK or MB_ICONERROR);
       Result := 0;
       exit;
    end;

  ZeroMemory( @dmScreenSettings, sizeof( DEVMODE ) );
  dmScreenSettings.dmSize := sizeof( DEVMODE );
  dmScreenSettings.dmPelsWidth  := 640;								// Width
  dmScreenSettings.dmPelsHeight := 480;								// Height
  dmScreenSettings.dmFields     := DM_PELSWIDTH or DM_PELSHEIGHT;		// Color Depth
  ChangeDisplaySettings(dmScreenSettings, CDS_FULLSCREEN); // Switch To Fullscreen Mode

  ShowWindow(h_Wnd, SW_SHOW);
  UpdateWindow(h_Wnd);
  SetFocus(h_Wnd);

  while ( true ) do begin
     // Process All Messages
    while (PeekMessage(msg, 0, 0, 0, PM_NOREMOVE)) do begin
      if (GetMessage(msg, 0, 0, 0)) then begin
        TranslateMessage(msg);
        DispatchMessage(msg);
        end
      else begin
        Result := 1;
        exit;
      end;
    end;

    DrawGLScene();
    SwapBuffers(h_DC);

    if (keys[VK_ESCAPE])
    then SendMessage(h_Wnd,WM_CLOSE,0,0)
    else ProcessKeys;
  end;//while
end;

{ tObjet }

constructor tObjet.Create(const FileName:string);
begin
  inherited Create;
  LoadFromFile(FileName);
end;

constructor tObjet.Create;
begin
  inherited Create;
end;

destructor tObjet.Destroy;
begin
  Points:=nil;
  inherited Destroy;
end;

function tObjet.LoadFromFile(const Filename: string):boolean;
var
  F			:TextFile;
  i			:integer;
  rx,ry,rz	:GLfloat;
  tmp		:string;
begin
  Result:=False;
  AssignFile(F,FileName);
  Reset(F);
  Readln(F,tmp);
  verts:=StrToInt(Copy(tmp,10,length(tmp)-9));
  SetLength(Points,verts);
  try
    for i:=0 to verts-1 do begin
      readln(F,rx,ry,rz);
      Points[i].x:=rx;
      Points[i].y:=ry;
      Points[i].z:=rz;
    end;//for
  finally
    CloseFile(F);
  end;//finally
  Result:=True;
end;

begin
  WinMain( hInstance, hPrevInst, CmdLine, CmdShow );
end.
